@@ -13,14 +13,15 @@
  *	one interpreter will be used to add/remove mounts and volumes,
  *	it does cope with multiple interpreters in multiple threads.
  *	
- * Copyright (c) 2001-2003 Vince Darley.
+ * Copyright (c) 2001-2004 Vince Darley.
  * 
  * See the file "license.terms" for information on usage and redistribution
  * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
  */
 
 #include <tcl.h>
-/* Required to access the 'stat' structure fields */
+/* Required to access the 'stat' structure fields, and TclInExit() */
+#include "tclInt.h"
 #include "tclPort.h"
 
 /*
@@ -33,6 +34,10 @@
 #define TCL_STORAGE_CLASS DLLEXPORT
 #endif /* BUILD_vfs */
 
+#ifndef TCL_GLOB_TYPE_MOUNT
+#define TCL_GLOB_TYPE_MOUNT		(1<<7)
+#endif
+
 /*
  * Only the _Init function is exported.
  */
@@ -277,7 +282,7 @@ static void            Vfs_RegisterWithInterp _ANSI_ARGS_((Tcl_Interp*));
 
 /* Some private helper procedures */
 
-static VfsNativeRep*   VfsGetNativePath(Tcl_Obj* pathObjPtr);
+static VfsNativeRep*   VfsGetNativePath(Tcl_Obj* pathPtr);
 static Tcl_CloseProc   VfsCloseProc;
 static void            VfsExitProc(ClientData clientData);
 static Tcl_Obj*	       VfsFullyNormalizePath(Tcl_Interp *interp, 
@@ -326,6 +331,7 @@ Vfs_Init(interp)
     if (Tcl_PkgRequire(interp, "Tcl", "8.4", 0) == NULL) {
 	return TCL_ERROR;
     }
+    
     /* 
      * Safe interpreters are not allowed to modify the filesystem!
      * (Since those modifications will affect other interpreters).
@@ -333,7 +339,12 @@ Vfs_Init(interp)
     if (Tcl_IsSafe(interp)) {
         return TCL_ERROR;
     }
-    if (Tcl_PkgProvide(interp, "vfs", "1.2.1") == TCL_ERROR) {
+
+#ifndef PACKAGE_VERSION
+    /* keep in sync with actual version */
+#define PACKAGE_VERSION "1.3"
+#endif
+    if (Tcl_PkgProvide(interp, "vfs", PACKAGE_VERSION) == TCL_ERROR) {
         return TCL_ERROR;
     }
 
@@ -472,7 +483,7 @@ Vfs_AddMount(mountPoint, isVolume, interp, mountCmd)
 	return TCL_ERROR;
     }
     strRep = Tcl_GetStringFromObj(mountPoint, &len);
-    newMount->mountPoint = (char*) ckalloc(1+len);
+    newMount->mountPoint = (char*) ckalloc(1+(unsigned)len);
     newMount->mountLen = len;
     
     if (newMount->mountPoint == NULL) {
@@ -625,7 +636,7 @@ Vfs_FindMount(pathMount, mountLen)
     mountIter = listOfMounts;
     while (mountIter != NULL) {
 	if (mountIter->mountLen == mountLen && 
-	  !strncmp(mountIter->mountPoint, mountStr, mountLen)) {
+	  !strncmp(mountIter->mountPoint, mountStr, (size_t)mountLen)) {
 	    Vfs_InterpCmd *ret = &mountIter->interpCmd;
 	    Tcl_MutexUnlock(&vfsMountsMutex);
 	    return ret;
@@ -666,7 +677,6 @@ Vfs_ListMounts(void)
     Tcl_MutexUnlock(&vfsMountsMutex);
     return res;
 }
-
 
 /*
  *----------------------------------------------------------------------
@@ -717,7 +727,6 @@ VfsFilesystemObjCmd(dummy, interp, objc, objv)
 
     switch ((enum options) index) {
 	case VFS_INTERNAL_ERROR: {
-	    int posixError = -1;
 	    if (objc > 3) {
 		Tcl_WrongNumArgs(interp, 2, objv, "?script?");
 		return TCL_ERROR;
@@ -732,11 +741,11 @@ VfsFilesystemObjCmd(dummy, interp, objc, objv)
 	    } else {
 		/* Set the script */
 		int len;
-		CONST char* str = Tcl_GetStringFromObj(objv[2],&len);
 		Tcl_MutexLock(&internalErrorMutex);
 		if (internalErrorScript != NULL) {
 		    Tcl_DecrRefCount(internalErrorScript);
 		}
+		Tcl_GetStringFromObj(objv[2], &len);
 		if (len == 0) {
 		    /* Clear our script */
 		    internalErrorScript = NULL;
@@ -1044,8 +1053,8 @@ VfsPathInFilesystem(Tcl_Obj *pathPtr, ClientData *clientDataPtr) {
  * path object, or NULL if no such representation exists.
  */
 static VfsNativeRep* 
-VfsGetNativePath(Tcl_Obj* pathObjPtr) {
-    return (VfsNativeRep*) Tcl_FSGetInternalRep(pathObjPtr, &vfsFilesystem);
+VfsGetNativePath(Tcl_Obj* pathPtr) {
+    return (VfsNativeRep*) Tcl_FSGetInternalRep(pathPtr, &vfsFilesystem);
 }
 
 static void 
@@ -1079,7 +1088,7 @@ VfsFilesystemPathType(Tcl_Obj *pathPtr) {
 }
 
 static Tcl_Obj*
-VfsFilesystemSeparator(Tcl_Obj* pathObjPtr) {
+VfsFilesystemSeparator(Tcl_Obj* pathPtr) {
     char sep=VFS_SEPARATOR;
     return Tcl_NewStringObj(&sep,1);
 }
@@ -1431,6 +1440,17 @@ VfsCloseProc(ClientData clientData) {
      * the Tcl code to use the channel's string-name).
      */
     Tcl_RegisterChannel(interp, chan);
+
+    if (!(Tcl_GetChannelMode(chan) & TCL_READABLE)) {
+	/* 
+	 * We need to make this channel readable, since tclvfs
+	 * documents that close callbacks are allowed to read
+	 * from the channels we create.
+	 */
+	
+	/* Currently if we reach here we have a bug */
+    }
+    
     returnVal = Tcl_EvalObjEx(interp, channelRet->closeCallback, 
 		  TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
     if (returnVal != TCL_OK) {
@@ -1460,50 +1480,85 @@ VfsMatchInDirectory(
     Tcl_GlobTypeData *types)	/* Object containing list of acceptable types.
 				 * May be NULL. */
 {
-    Tcl_Obj *mountCmd = NULL;
-    Tcl_SavedResult savedResult;
-    int returnVal;
-    Tcl_Interp* interp;
-    int type = 0;
-    Tcl_Obj *vfsResultPtr = NULL;
-    
-    mountCmd = VfsBuildCommandForPath(&interp, "matchindirectory", dirPtr);
-    if (mountCmd == NULL) {
-	return -1;
-    }
+    if ((types != NULL) && (types->type & TCL_GLOB_TYPE_MOUNT)) {
+	VfsMount *mountIter;
+	int len;
+	CONST char *prefix;
 
-    if (types != NULL) {
-	type = types->type;
-    }
-
-    if (pattern == NULL) {
-	Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewObj());
-    } else {
-	Tcl_ListObjAppendElement(interp, mountCmd, 
-				 Tcl_NewStringObj(pattern,-1));
-    }
-    Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewIntObj(type));
-    Tcl_SaveResult(interp, &savedResult);
-    /* Now we execute this mount point's callback. */
-    returnVal = Tcl_EvalObjEx(interp, mountCmd, 
-			      TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
-    if (returnVal != -1) {
-	vfsResultPtr = Tcl_DuplicateObj(Tcl_GetObjResult(interp));
-    }
-    Tcl_RestoreResult(interp, &savedResult);
-    Tcl_DecrRefCount(mountCmd);
-
-    if (vfsResultPtr != NULL) {
-	if (returnVal == TCL_OK) {
-	    Tcl_IncrRefCount(vfsResultPtr);
-	    Tcl_ListObjAppendList(cmdInterp, returnPtr, vfsResultPtr);
-	    Tcl_DecrRefCount(vfsResultPtr);
-	} else {
-	    Tcl_SetObjResult(cmdInterp, vfsResultPtr);
+	prefix = Tcl_GetStringFromObj(Tcl_FSGetNormalizedPath(NULL, dirPtr), 
+				      &len);
+	if (prefix[len-1] == '/') {
+	    /* 
+	     * It's a root directory; we must subtract one for
+	     * our comparisons below
+	     */
+	    len--;
 	}
+
+	Tcl_MutexLock(&vfsMountsMutex);
+
+	/* Build list of mounts */
+	mountIter = listOfMounts;
+	while (mountIter != NULL) {
+	    if (mountIter->mountLen > (len+1) 
+		&& !strncmp(mountIter->mountPoint, prefix, (size_t)len) 
+		&& mountIter->mountPoint[len] == '/'
+		&& strchr(mountIter->mountPoint+len+1, '/') == NULL
+		&& Tcl_StringCaseMatch(mountIter->mountPoint+len+1, 
+				       pattern, 0)) {
+		Tcl_Obj* mount = Tcl_NewStringObj(mountIter->mountPoint, 
+						  mountIter->mountLen);
+		Tcl_ListObjAppendElement(NULL, returnPtr, mount);
+	    }
+	    mountIter = mountIter->nextMount;
+	}
+	Tcl_MutexUnlock(&vfsMountsMutex);
+	return TCL_OK;
+    } else {
+	Tcl_Obj *mountCmd = NULL;
+	Tcl_SavedResult savedResult;
+	int returnVal;
+	Tcl_Interp* interp;
+	int type = 0;
+	Tcl_Obj *vfsResultPtr = NULL;
+	
+	mountCmd = VfsBuildCommandForPath(&interp, "matchindirectory", dirPtr);
+	if (mountCmd == NULL) {
+	    return -1;
+	}
+
+	if (types != NULL) {
+	    type = types->type;
+	}
+
+	if (pattern == NULL) {
+	    Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewObj());
+	} else {
+	    Tcl_ListObjAppendElement(interp, mountCmd, 
+				     Tcl_NewStringObj(pattern,-1));
+	}
+	Tcl_ListObjAppendElement(interp, mountCmd, Tcl_NewIntObj(type));
+	Tcl_SaveResult(interp, &savedResult);
+	/* Now we execute this mount point's callback. */
+	returnVal = Tcl_EvalObjEx(interp, mountCmd, 
+				  TCL_EVAL_GLOBAL | TCL_EVAL_DIRECT);
+	if (returnVal != -1) {
+	    vfsResultPtr = Tcl_DuplicateObj(Tcl_GetObjResult(interp));
+	}
+	Tcl_RestoreResult(interp, &savedResult);
+	Tcl_DecrRefCount(mountCmd);
+
+	if (vfsResultPtr != NULL) {
+	    if (returnVal == TCL_OK) {
+		Tcl_IncrRefCount(vfsResultPtr);
+		Tcl_ListObjAppendList(cmdInterp, returnPtr, vfsResultPtr);
+		Tcl_DecrRefCount(vfsResultPtr);
+	    } else {
+		Tcl_SetObjResult(cmdInterp, vfsResultPtr);
+	    }
+	}
+	return returnVal;
     }
-    
-    return returnVal;
 }
 
 static int
@@ -1669,14 +1724,20 @@ VfsFileAttrsGet(cmdInterp, index, pathPtr, objPtrRef)
 	     */
 	} else {
 	    /* Leave error message in correct interp */
-	    Tcl_SetObjResult(cmdInterp, *objPtrRef);
+	    if (cmdInterp != NULL) {
+		Tcl_SetObjResult(cmdInterp, *objPtrRef);
+	    } else {
+		Tcl_DecrRefCount(*objPtrRef);
+	    }
 	    *objPtrRef = NULL;
 	}
     } else {
-	Tcl_ResetResult(cmdInterp);
-	Tcl_AppendResult(cmdInterp, "couldn't read attributes for \"", 
-			 Tcl_GetString(pathPtr), "\": ",
-			 Tcl_PosixError(cmdInterp), (char *) NULL);
+	if (cmdInterp != NULL) {
+	    Tcl_ResetResult(cmdInterp);
+	    Tcl_AppendResult(cmdInterp, "couldn't read attributes for \"", 
+			     Tcl_GetString(pathPtr), "\": ",
+			     Tcl_PosixError(cmdInterp), (char *) NULL);
+	}
     }
     
     return returnVal;
@@ -1713,19 +1774,22 @@ VfsFileAttrsSet(cmdInterp, index, pathPtr, objPtr)
     Tcl_RestoreResult(interp, &savedResult);
     Tcl_DecrRefCount(mountCmd);
     
-    if (returnVal == -1) {
-	Tcl_ResetResult(cmdInterp);
-	Tcl_AppendResult(cmdInterp, "couldn't set attributes for \"", 
-			 Tcl_GetString(pathPtr), "\": ",
-			 Tcl_PosixError(cmdInterp), (char *) NULL);
+    if (cmdInterp != NULL) {
+	if (returnVal == -1) {
+	    Tcl_ResetResult(cmdInterp);
+	    Tcl_AppendResult(cmdInterp, "couldn't set attributes for \"", 
+			     Tcl_GetString(pathPtr), "\": ",
+			     Tcl_PosixError(cmdInterp), (char *) NULL);
+	} else if (errorPtr != NULL) {
+	    /* 
+	     * Leave error message in correct interp, errorPtr was
+	     * duplicated above, in case of threading issues.
+	     */
+	    Tcl_SetObjResult(cmdInterp, errorPtr);
+	}
     } else if (errorPtr != NULL) {
-	/* 
-	 * Leave error message in correct interp, errorPtr was
-	 * duplicated above, in case of threading issues.
-	 */
-	Tcl_SetObjResult(cmdInterp, errorPtr);
+	Tcl_DecrRefCount(errorPtr);
     }
-    
     return returnVal;
 }
 
